// THIS IS AN INITIAL IMPLEMENTATION OF MATCHERCELLS
// YOU NEED A JAVASCRIPT INTERPRETER (eg. ringojs, nodejs) TO EXECUTE THIS FILE


//HELPER FUNCTIONS ==============================================
var sleep = function(milliSeconds) {
    var startTime = getTime();
    while (getTime() < startTime + milliSeconds);
};

var getTime = function() {
    return (new Date()).getTime();
};

var length = function (array){
    return array.length;
};

var map = function (array, fun, arg){
    return array.map(function (el){
        return fun(el, arg);
    });
};

var difference = function (array1, array2){
    return array1.filter(function (element1){
        return !array2.some(function (element2){
            return element1 === element2;
        });
    });
};

var removeDuplicates = function (array){
    var newArray = [];
    for(var i = 0; i < array.length; ++i){
        var duplicated = false;
        for(var j = 0; j < newArray.length; ++j){
            if(array[i] == newArray[j]){
                duplicated = true;
                break;
            }
        }
        if(!duplicated){
            newArray.push(array[i]);
        }
    }
    return newArray;
};

var append = function (array1, array2){
    return array1.concat(array2);
};

var lifeTime = function(cell) {
    return getTime() - cell.traceLifeTime();
};

var onlyMatchCells = function (cells){
    for(var i = 0; i < cells.length; ++i){
        if(!cells[i].isMatch()){
            return false;
        }
    }
    return true;
};

var getCreators = function (newCells,cells) {
    return parents(difference(newCells, cells));
};

var parents = function (cells){
    return cells.map(function (cell){
        return cell.parent();
    });
};

var createSeed = function (sequence){
    return new TimeCell(sequence, null);
};

//Weaving of a stateful aspect and the evaluation of a cell ====================
var weaving = function (sAsp, jps){
    print("weaving of " + jps);
    for(var i = 0; i < jps.length; ++i){
        print("jp:" + jps[i]);

        //patch to emulate time
        if (jps[i] == "t") {
            sleep(4000);
        }

        var matches = sAsp.eval(jps[i]);
        for(var j = 1; j <= matches; ++j){
            sAsp.advice();
        }
    }
};

var StatefulAspect = function (sequence, advice, rule){
    this.advice = advice;
    this.sequence = sequence;
    this.rule = rule;

    var cells = [createSeed(sequence)];

    this.eval = function (jp){
        cells = this.rule(cells, jp);
        var matches = this.match();
        cells = cells.filter(function (cell){
            return !cell.isMatch();
        });
        return matches;
    };

    this.match = function (){
        var cont = 0;
        cells.forEach(function (cell){
            cont += cell.isMatch() ? 1 : 0;
        });
        return cont;
    };
};

var adv = function (){
    print("M");
};

var Cell = function (sequence, parent){
    print("hola");

    this.eval = function (jp){
        return sequence[0] == jp ? new Cell(sequence.substr(1), this) : this;
    };

    this.isMatch = function (){
        return sequence == "";
    };

    this.parent = function (){
        return parent;
    };

    this.toString = function (){
        return sequence;
    };
};

var TimeCell = function (sequence, parent, traceLifeTime){
    this.eval = function (jp){
        if (sequence[0] == jp) {
            // simplify!!!
            var time = parent == null? getTime() : traceLifeTime;
            return new TimeCell(sequence.substr(1), this ,time);
        }
        return this;
    };

    this.traceLifeTime = function() {
      return traceLifeTime || 0;
    };

    this.isMatch = function (){
        return sequence == "";
    };

    this.parent = function (){
        return parent;
    };

    this.toString = function (){
        return sequence;
    };
};


var react = function (cell, jp){
    return cell.eval(jp);
};

//RULES =================================
var applyReaction = function (cells, jp){
    return removeDuplicates(append(map(cells, react, jp), cells));
};

var killCreators = function (rule){
    return function (cells, jp){
        var newCells = rule.call(this,cells, jp);
        return difference(newCells, getCreators(newCells,cells));
    }
};

var addSeed = function (rule){
    return function (cells, jp){
        var newCells = rule.call(this, cells, jp);
        return length(newCells) == 0 || onlyMatchCells(newCells) ?
               append(newCells, [createSeed(this.sequence)]) : newCells;
    }
};

var traceLifeTime = function (rule, time){
    return function (cells, jp){
        var newCells = rule.call(this,cells,jp);
        return newCells.filter(function(cell) {
           return lifeTime(cell) <= time;
        });
    }
};

var cellsWithDifferentInformation = function (rule){
    return function (cells, jp){
        var newCells = rule(cells, jp);
        return newCells.filter(newCells, function (cell){
            return hasDifferentBindings(cell, newCells);
        });
    }
};


//SOME TESTS================================
weaving(new StatefulAspect("ab", adv, applyReaction), "aab");
weaving(new StatefulAspect("ab", adv, killCreators(applyReaction)), "aab");
weaving(new StatefulAspect("ab", adv, addSeed(killCreators(applyReaction))), "aabab");
weaving(new StatefulAspect("ab", adv, addSeed(traceLifeTime(killCreators(applyReaction), 2000))), "atab");
